#!/usr/bin/env python3

import sys
import time
import datetime
import re
import pprint
import logging
import json
import mmap
import lib_path
import lib
import dphysldap

update_every = 1
log_level = logging.DEBUG
log_datefmt = '%Y-%m-%d %H:%M:%S'
log_path = '/var/log/ldap/'
log_file_log = log_path + 'deltalogparse.log'
log_file_deltalog = log_path + 'delta.log'
search_base = 'cn=deltalog'
object_classes = '(|(objectClass=auditModify)(objectClass=auditAdd)(objectClass=auditDelete)(objectClass=auditModRDN))'

log = logging.getLogger('log')
log.setLevel(log_level)
log_fh = logging.FileHandler(log_file_log)
log_fh.setLevel(log_level)
log_formatter = logging.Formatter(fmt='{asctime} {levelname}: {message}', style='{', datefmt=log_datefmt)
log_fh.setFormatter(log_formatter)
log.addHandler(log_fh)

deltalog = logging.getLogger('deltalog')
deltalog.setLevel(logging.INFO)
deltalog_fh = logging.FileHandler(log_file_deltalog)
deltalog_fh.setLevel(logging.INFO)
deltalog_formatter = logging.Formatter(fmt='{asctime}: {message}', style='{', datefmt=log_datefmt)
deltalog_fh.setFormatter(deltalog_formatter)
deltalog.addHandler(deltalog_fh)

rgx_skip = re.compile(
    r'^('
    r'heartbeatTimestamp|'
    r'modifyTimestamp|'
    r'modifiersName|'
    r'lastUse.*?|'
    r'(context|entry)CSN'
    r'):')

rgx_filter = re.compile(
    r'^('
    r'heartbeatTimestamp|'
    r'lastUse.*?|'
    r'(context|entry)CSN'
    r'):')


def getlastline(fname):
    """Using mmap to return a copy of the last line of a file"""
    with open(fname) as source:
        mapping = mmap.mmap(source.fileno(), 0, prot=mmap.PROT_READ)
    return mapping[mapping.rfind(b'\n', 0, -1)+1:]


def is_skipped(entry):
    """Check if at least on of the reqMod attributes is interesting"""
    for attribute in entry['attributes']['reqMod']:
        if not rgx_skip.search(attribute):
            log.info('interesting attribute: {}'.format(attribute))
            return False

    return True


def filtered(entry):
    """Filter out attributes not of interest"""
    if 'reqMod' in  entry['attributes']:
        for attribute in entry['attributes']['reqMod']:
            if rgx_filter.search(attribute):
                log.info('filter attribute: {}'.format(attribute))
                entry['attributes']['reqMod'].remove(attribute)

    del entry['raw_dn']
    del entry['raw_attributes']

    return entry


def log_action(action, entry):
    """Log an action on a req"""
    req_type = entry['attributes']['reqType'][0]
    req_dn = entry['attributes']['reqDN'][0]
    if not action == 'skip':
        log.debug('entry:')
        log.debug(pprint.pformat(entry, indent=1))
    log.info('{}: {} {}'.format(action, req_type, req_dn))


def write(entry):
    """Write a req to file"""
    req_type = entry['attributes']['reqType'][0]
    req_dn = entry['attributes']['reqDN'][0]

    entry['attributes'] = dict(entry['attributes'])

    # debug
    entry['raw_attributes'] = dict(entry['raw_attributes'])
    log.debug('original entry:')
    log.debug(pprint.pformat(entry, indent=1))

    filtered_entry = filtered(entry)

    log_action('write', filtered_entry)
    deltalog.info('{}'.format(json.dumps(filtered_entry)))


def sleep(start_time):
    """After hard work, take a nap for the rest of the second"""
    current_time = time.perf_counter()
    elapsed_time = current_time - start_time
    sleep_time = update_every - elapsed_time

    log.info('runtime {0:.3f}s, sleeping {1:.3f}s'.format(elapsed_time, sleep_time))

    if sleep_time > 0:
        time.sleep(sleep_time)


def main():
    """Connect to slapd socket, search accesslog, write interesting changes"""
    slapd = dphysldap.Slapd()
    try:
        slapd.connect()
    except:
        log.exception('error: failed to connect! trying to reconnect...')
        if slapd.reconnect():
            log.error('connected to slapd')

    try:
        log.info('trying to read last req from {}'.format(log_file_deltalog))
        last_line = getlastline(log_file_deltalog).decode('utf-8')
        req_json = last_line.split(': ', maxsplit=1)[1]
        log.debug('last line: {}'.format(req_json))
        req = json.loads(req_json)
        log.debug(pprint.pformat(req, indent=1))
        req_start = req['attributes']['reqStart'][0]
        log.info('continuing from last logged req: {}'.format(req_start))
    except:
        log.exception('error: while trying to retrieve last req')
        req_start = datetime.datetime.utcnow().strftime('%Y%m%d%H%M%S.%fZ')
        log.info('using current timestamp: {}'.format(req_start))

    while True:
        start_time = time.perf_counter()
        
        search_filter=''.join(['(&', object_classes, '(reqStart>=', req_start, ')',')'])

        try:
            response = slapd.search(search_base, search_filter)
        except:
            log.exception('error: reading socket! trying to reconnect...')
            if slapd.reconnect():
                log.error('socket reconnected.')
                continue

        if response:
            for entry in slapd.response():
                if entry['attributes']['reqStart'][0] == req_start:
                    continue

                req_start = entry['attributes']['reqStart'][0]
                req_type = entry['attributes']['reqType'][0]
                req_dn = entry['attributes']['reqDN'][0]

                log.info('processing: {}'.format(entry['dn']))

                if req_type == 'modify':
                    if is_skipped(entry):
                        log_action('skip', entry)
                        continue

                write(entry)

        sleep(start_time)


if __name__ == '__main__':
    main()
    sys.exit(0)
